#!/usr/bin/perl

use strict;
use File::Copy;
use POSIX qw();
use Data::Dumper;
use URI;
use File::Basename;
use Net::FTP;
use Net::SCP::Expect;
use LWP::UserAgent;
use Unicode::MapUTF8 qw(to_utf8 from_utf8);
use MIME::Lite;
use Net::SMTP;
use Net::Domain;
use Config::Properties;
use Log::Log4perl qw(:easy);

##******************************************************************************
## Gobal configuration variables
##******************************************************************************

# Remove uninitialized warnings
no warnings 'uninitialized';

# Set the Root directory
my ($rootInstall) = "/var/lib/interldap/atransport-engine";

# Input files directory
my ($inputDir)  = "$rootInstall/input/";   # mettre un / a la fin !
# Waiting files directory
my ($fifosDir)  = "$rootInstall/fifos/";   # mettre un / a la fin !
# Well transmetted files directory
my ($doneDir)   = "$rootInstall/done/";    # mettre un / a la fin !
# Directory where to store files which could not be transmitted
my ($errorDir)  = "$rootInstall/error/";   # mettre un / a la fin !

# Time (in seconds) between each directory scanning
my $timeBetweenDistributions = 2;

# Time (in seconds) between sending two files
my $timeBetweenSending = 1;

# Please specify '-d' option to get debug messages
my $debug = 0;
if( exists( $ARGV[0] ) and $ARGV[0] eq '-d' ) {
  $debug = 1;
  Log::Log4perl->easy_init($DEBUG);
} else {
  Log::Log4perl->init("/etc/interldap/atransport-engine-log.conf");
}

# Get the Log4Perl Logger
my $logger = get_logger();

# Properties file
my $confDirectory = "/etc/interldap";
my $confFilename = "$confDirectory/atransport-engine.properties";

##******************************************************************************
## Property file usage
##******************************************************************************
open( PROPS, "<$confFilename" ) or die("open $confFilename : $!");
my $properties = new Config::Properties();
$properties->load(*PROPS);
my $treatment = $properties->getProperty( "export.treatment", '/bin/true' );
my $mailFromName = $properties->getProperty( "export.mail_from_name", 'Moteur d\'exportation' );
my $mailFromAdress = $properties->getProperty( "export.mail_from_adress", 'root@localhost' );
my $mailSubject = $properties->getProperty( "export.mail_subject", 'Transfert d\'un fichier' );
close(PROPS);
$properties = undef;

my $user = $ENV{USER};
my $host = Net::Domain::hostfqdn();
$mailFromAdress =~ s/%u/$user/g;
$mailFromAdress =~ s/%h/$host/g;
($user,$host) = ();

$logger->debug("treatment : $treatment\n");
$logger->debug("mailFromName : $mailFromName\n");
$logger->debug("mailFromAdress : $mailFromAdress\n");
$logger->debug("mailSubject : $mailSubject\n");

##******************************************************************************
## Other global variables
##******************************************************************************
my $tmpDir    = undef;

my %serverPid = ();
my @killedPid = ();

##******************************************************************************
## Function getInfosFromURI (raise exceptions)
## extracts information from URI
##******************************************************************************
sub getInfosFromURI
{
   my ($destinationString) = @_;
   my ($scheme, $user, $passwd, $host, $path) =
         ( $destinationString =~ m|^([^:]+)://([^:@/]+)(?::([^:@/]+))?@([^@/]*)(.*)$| );
   if( !defined( $path ) )
   {
      my ($server, $port, $email) =
            ( $destinationString =~ m|^mail://(?:([^:/]+)(?::([^:@/]+))?)?/(.*)$| );
      die("Bad format for URI : $destinationString") if( ! defined( $email ) );
      return { scheme=>'mail', server=>$server, port=>$port, email=>$email };
   }
   return { scheme=>$scheme, user=>$user, passwd=>$passwd, host=>$host, path=>$path };
}

##******************************************************************************
## Function sendHTTP (raise exceptions)
## send through HTTP
##******************************************************************************
sub sendHTTP
{
   my ($destinationInfos, $baseName, $tmpFileName) = @_;

   my $url = 'http://'.$destinationInfos->{user}.':'.$destinationInfos->{passwd}.'@'.$destinationInfos->{host}.$destinationInfos->{path};

   my $ua = LWP::UserAgent->new();
   my $res=$ua->post("$url",
       Content_Type => 'multipart/form-data',
       Content      => [ user       => $destinationInfos->{user},
                         password   => $destinationInfos->{passwd},
                         nomfichier => $baseName,
                         fichier    => [$tmpFileName],
                        ],
       ) or die("ua->post(): $!\n");

   die( $res->error_as_HTML() )
      if( ! $res->is_success() );
}

##******************************************************************************
## Function sendSCP (raise exceptions)
## Send through SCP
##******************************************************************************
sub sendSCP
{
   my ($destinationInfos, $baseName, $tmpFileName) = @_;

   my $pid = fork();
   die("$0: fork: $!") if( $pid == -1 );
   if( $pid )
   {  # Pere
      wait();
   }
   else
   {  # Fils
      open( STDIN, "</dev/null" );
      open( STDOUT, ">/dev/null" );
      open( STDERR, ">/dev/null" );
      {
         exec( "/usr/bin/scp",
               $tmpFileName,
               "$destinationInfos->{user}\@$destinationInfos->{host}:$destinationInfos->{path}" );
      }
      exit( 1 );
   }
}


#sub sendSCP
#{
#   my ($destinationInfos, $baseName, $tmpFileName) = @_;
#
#   my $scpe = Net::SCP::Expect->new( host     => $destinationInfos->{host},
#                                     user     => $destinationInfos->{user},
#                                     password => $destinationInfos->{passwd} )
#      or die( "Net::SCP::Expect->new(): $!" );
#
#   $scpe->scp( $tmpFileName, $destinationInfos->{path} )
#      or die("scpe->scp(): $!");
#}

##******************************************************************************
## Function sendFTP (raise exceptions)
## Send through FTP
##******************************************************************************
sub sendFTP
{
   my ($destinationInfos, $baseName, $tmpFileName) = @_;

   my $ftp = Net::FTP->new( $destinationInfos->{host}, Debug => 0, Passive => 1 )
      or die "Net::FTP->new(): $@";

   $ftp->login( $destinationInfos->{user}, $destinationInfos->{passwd} )
      or die("ftp->login(): ". $ftp->message());

   $ftp->cwd( $destinationInfos->{path} )
      or die("ftp->cwd(): ",$ftp->message());

   $ftp->binary()
      or die("ftp->binary(): ", $ftp->message());

   $ftp->put( $tmpFileName, $baseName )
      or die("ftp->put(): ", $ftp->message());

   $ftp->quit()
      or die("ftp->quit(): ", $ftp->message());
}

##******************************************************************************
## Function sendSMTP (raise exceptions)
## Send through SMTP
##******************************************************************************
sub sendSMTP
{
   my ($destinationInfos, $baseName, $tmpFileName) = @_;

   # Valeurs par d�faut :
   $destinationInfos->{server} ||= 'localhost';
   $destinationInfos->{port}   ||= '25';

   my $subject = $mailSubject;
   $subject =~ s/%f/$baseName/g;
   $logger->debug("subject : $subject\n");

   my $mime = MIME::Lite->new(
      From    => $mailFromName.'<'.$mailFromAdress.'>',
      To      => $destinationInfos->{email},
      Subject => $subject,
      Type    => 'multipart/mixed',
   );
   $mime->attach(
      Type       => 'application/xml',
      Encoding   => 'base64',
      Path       => $tmpFileName,
      Filename   => $baseName,
   );

   my $smtp = Net::SMTP->new( $destinationInfos->{server},
                              Port => $destinationInfos->{port} )
      or die("Net::SMTP->new(): $!");

   $smtp->mail($mailFromAdress)          or die("smtp->mail(): $!");
   $smtp->to($destinationInfos->{email}) or die("smtp->to(): $!");
   $smtp->data()                         or die("smtp->data(): $!");
   $smtp->datasend( $mime->as_string() ) or die("smtp->datasend(): $!");
   $smtp->dataend()                      or die("smtp->dataend(): $!");
   $smtp->quit()                         or die("smtp->quit(): $!");

}

##******************************************************************************
## Variable senderFunction
## Array of functions by protocol
##******************************************************************************
my $senderFunction =
{
   ftp  => \&sendFTP,
   http => \&sendHTTP,
   scp  => \&sendSCP,
   sftp => \&sendSCP,
   mail => \&sendSMTP,
};

##******************************************************************************
## Function extractDate
## use to sort files according to their name
##******************************************************************************
sub extractDate
{
   my ($file) = @_;
   my ($export,$date) = ( $file =~ m:^.*/([\w-]+?)(\d+)[^/]*$: );
   return int($date);
}

##******************************************************************************
## Function senderMain
## Send main function
##******************************************************************************
sub senderMain
{
   my ($export) = @_;
   $0 .= " sender for $export"; # pour le voir dans ps

   my %waitingFiles = ();

   while(1)
   {
      LABELFICHIER: foreach my $file (sort {extractDate($a)<=>extractDate($b)} glob($fifosDir.$export.'/*.hdr'))
      {
         my $baseName = "";
         my $tmpFileName = "";
         my $dataFile = $file; 

         eval # Gestion des exceptions
         {

            # --------------------------------------------------------
            # Read the headers file
            open(FILE, $file) or die("open($file) : $!");

            my $ligne = <FILE>;
            my ($extension) = ( $ligne =~ m/^Extension\s*:\s*(.+)$/ );
            defined($extension) or die("Error: no extension line in $file");
            $dataFile =~ s/hdr$/$extension/;
            $baseName = basename($dataFile);

            # Get the encoding
            $ligne = <FILE>;
            my ($encodingString) = ( $ligne =~ m/^Encoding\s*:\s*(.+)$/ );
            defined($encodingString) or die("Error: no encoding line in $file");
            my ($carriageReturn,$encoding) = ( $encodingString =~ m/^\{(.)\}(.*)$/ );

            # Get the destination
            $ligne = <FILE>;
            my ($destinationString) = ( $ligne =~ m/^URI\s*:\s*(.+)$/ );

            # If no destination, it's because it was the last one : done
            if( ! defined($destinationString) )
            {
               close FILE;
               move( $file, $doneDir ) or $logger->warn("Unable to move $file to $doneDir \n");
               move( $dataFile, $doneDir ) or $logger->warn("Unable to move $dataFile to $doneDir \n");
            }
            else
            {

               # --------------------------------------------------------
               # Write file
               open(FILE2, ">$file.tmp") or die("open($file) : $!");
               print FILE2 "Extension: $extension\n";
               print FILE2 "Encoding: $encodingString\n";
               while( $ligne = <FILE> )
               {
                  print FILE2 $ligne;
               }
               close FILE;
               close FILE2;
               rename( "$file.tmp", $file ) or die("rename( $file.tmp, $file ): $!");

               # --------------------------------------------------------
               # Manage encoding and destination system
               open(FILE, $dataFile) or die("open($file) : $!");
               my $content;
               {
                  local $/ = undef;
                  $content = <FILE>;
               }
               close(FILE);

               $content = from_utf8({-string=>$content,-charset=>'ISO-8859-1'}); # Obligatoire pour
               $content = to_utf8  ({-string=>$content,-charset=>'ISO-8859-1'}); # avoir de l'UTF
               if( $encoding !~ /^UTF/ )
               {
                  $content = from_utf8({-string=>$content,-charset=>'ISO-8859-1'});
               }
               if( $carriageReturn eq 'W' )
               {
                  $content =~ s/\n/\r\n/g;
               }

               # Modified copy of in $tmpDir
               $tmpFileName = $tmpDir.'/'.$baseName;
               open( TMPFILE, ">$tmpFileName" ) or die("open $tmpFileName: $!");
               binmode TMPFILE;
               print( TMPFILE "$content" );
               close( TMPFILE );

               # --------------------------------------------------------
               # Manage execution
               my $execLine = $treatment;
               $execLine =~ s/%f/$tmpFileName/g;
               $logger->debug("system $execLine\n");
               {
                  local $SIG{CHLD} = 'DEFAULT';
                  system($execLine);
               }

               my $destinationInfos = getInfosFromURI( $destinationString );
               my $scheme = $destinationInfos->{scheme};
               die( "Unknown scheme '$scheme' file $file\n" )
                  if( ! exists( $senderFunction->{$scheme} ) );
               $senderFunction->{$scheme}->( $destinationInfos, $baseName, $tmpFileName );

            }

         }; # Evaluation end of the exceptions management

         if( $@ )  # If there is an exception
         {
            $logger->debug("Exception : $file : $@\n");
            if( ! exists( $waitingFiles{$baseName} ) )
            {  # First error :
               $waitingFiles{$baseName} = 1;
            }
            else
            {  # Second error :
               delete( $waitingFiles{$baseName} );
               move( $file, $errorDir ) or $logger->warn("Unable to move $file to $errorDir");
               move( $dataFile, $errorDir ) or $logger->warn("Unable to move $dataFile to $errorDir");
            }
         }

         unlink( $tmpFileName ) if( defined($tmpFileName) );

      }

      sleep( $timeBetweenSending );
   }
}

##******************************************************************************
## Function senderCreate
## Used to create a send based on the specified export
##******************************************************************************
sub senderCreate
{
   my ($export) = @_;

   my $pid = fork();
   die("$0: fork: $!") if( $pid == -1 );
   if( $pid )
   {  # Pere
      $serverPid{$pid} = $export;
   }
   else
   {  # Fils
      $SIG{CHLD} = 'IGNORE';
      senderMain( $export );
      die("$0: should not be here");
   }
}

##******************************************************************************
## Function recreateKilledSenders
## Recreate killed senders
##******************************************************************************
sub recreateKilledSenders
{
   my @killedPidCopy = @killedPid;
   foreach my $pid (@killedPidCopy)
   {
      my $export = $serverPid{$pid};
      delete $serverPid{$pid};
      senderCreate( $export );
      @killedPid = grep { $_ != $pid } @killedPid;
   }
}

##******************************************************************************
## Function fileDistribution
## Manage files distribution in fifos directories
## Call senderCreate if a new destination is detected
##******************************************************************************
sub fileDistribution
{
   my ($inputDir,$fifosDir) = @_;

   # Lecture des fichier :
   opendir(DIR, $inputDir) or die "opendir $inputDir: $!";
   my @inputFiles = grep( m/\.hdr$/, grep {-f} map {$inputDir.$_} readdir(DIR));
   closedir DIR;

   # Traitement des fichiers :
   foreach my $file ( @inputFiles )
   {
      if( my ($export,$date) = ( $file =~ m:^.*/([\w-]+?)(\d+)[^/]*$: ) )
      {
         $logger->debug("$file $export $date\n");

         if( ! -d $fifosDir.$export )
         {
            if( ! mkdir( $fifosDir.$export ) ) {
	       $logger->warn("$0: mkdir $fifosDir.$export: $! \n");
	    }
            senderCreate( $export );
         }

         # Look for the name in the data file
         my $dataFile = $file;
         if( open(FILE, $file) ) {
            my $ligne = <FILE>;
            my ($extension) = ( $ligne =~ m/(\S+)$/ );
            $dataFile =~ s/hdr$/$extension/;
            close FILE;
         } else {
            $logger->warn("$0: $file: $!\n");
            move( $file, $errorDir ) or $logger->warn("Unable to move $file to $errorDir !");
         }

         if( ! move($file,$fifosDir.$export) ) {
            $logger->warn("$0: move($file,$fifosDir.$export): $!\n");
         } elsif( ! move($dataFile,$fifosDir.$export) ) {
            $logger->warn("$0: move($dataFile,$fifosDir.$export): $!\n");
         }

      }
      else
      {
         $logger->warn("$0: $file: Bad name format\n");
         move( $file, $errorDir ) or $logger->warn("$0: move( $file, $errorDir ): $!\n");;
      }
   }
}

##******************************************************************************
## Function sigChildHandler
## Callet when a SIGCHILD signal is raised
##******************************************************************************
sub sigChildHandler
{
   $SIG{CHLD} = \&sigChildHandler;

   my $child;
   while( ( $child = waitpid(-1,&POSIX::WNOHANG) ) > 0 )
   {
      push @killedPid, $child;
   }
}
$SIG{CHLD} = \&sigChildHandler;

##******************************************************************************
## Function sigIntHandler
## Before exiting the process, kill all child
##******************************************************************************
sub sigIntHandler
{
   $logger->debug("Killing all sub process ...\n");
   kill 'INT' => keys %serverPid;
   rmdir( $tmpDir );
   exit(0);
}
$SIG{INT} = \&sigIntHandler;
$SIG{TERM} = \&sigIntHandler;


##******************************************************************************
## main
##******************************************************************************

#print "Now backgrounding $0 process...\n";
#fork and exit;

# Create the temporary directory
$tmpDir = POSIX::tmpnam();
mkdir( $tmpDir ) or die("mkdir $tmpDir: $!");

# Create senders for existing directories
opendir(DIR, $fifosDir) or die "opendir $fifosDir: $!";
my @exportDir = grep { !/^\./ && -d $fifosDir.$_ } readdir(DIR);
closedir DIR;
map { senderCreate( $_ ) } @exportDir;

# Wait for new files
while(1)
{
   fileDistribution( $inputDir, $fifosDir);
   recreateKilledSenders();
   sleep( $timeBetweenDistributions );
}

##******************************************************************************
